/*
* Copyright (C) 2011 Sunil Mohan Adapa <sunil@medhas.org>.
* Copyright (C) 2011 O S K Chaitanya <oskchaitanya@gmail.com>.
*
* Author: Sunil Mohan Adapa <sunil@medhas.org>
*         O S K Chaitanya <oskchaitanya@gmail.com>
*
* This file is part of GNOME Nonogram.
*
* GNOME Nonogram is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* GNOME Nonogram is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with GNOME Nonogram. If not, see <http://www.gnu.org/licenses/>.
*/

const Lang = imports.lang;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;

const Config = imports.config;
const NonogramIO = imports.nonogramio;

GameWidget = new GType(_GameWidget = {
    parent: Gtk.Table.type,
    name: "GameWidget",
    properties: [
        {
            name: "puzzleFile",
            type: GObject.TYPE_STRING,
            default_value: null,
            flags: GObject.ParamFlags.READABLE
                 | GObject.ParamFlags.WRITABLE
                 | GObject.ParamFlags.CONSTRUCT
        }
    ],

    _BACKGROUND_COLOR: 0xffffffff,
    _NO_COLOR: -1,

    _solutionPuzzlePixels: null,
    _numRows: 0,
    _numColumns: 0,

    _rowClueTable: null,
    _columnClueTable: null,
    _pixelGrid: null,
    _correctSolutionDialog: null,
    _incorrectSolutionDialog: null,
    _incorrectSolutionDialogShown: false,

    class_init: function(klass, prototype) {
        prototype._BACKGROUND_COLOR = _GameWidget._BACKGROUND_COLOR;
        prototype._NO_COLOR = _GameWidget._NO_COLOR;

        prototype._solutionPuzzlePixels = _GameWidget._solutionPuzzlePixels;
        prototype._numRows = _GameWidget._numRows;
        prototype._numColumns = _GameWidget._numColumns;

        prototype._rowClueTable = _GameWidget._rowClueTable;
        prototype._columnClueTable = _GameWidget._columnClueTable;
        prototype._pixelGrid = _GameWidget._pixelGrid;
        prototype._correctSolutionDialog = _GameWidget._correctSolutionDialog;
        prototype._incorrectSolutionDialog =
            _GameWidget._incorrectSolutionDialog;
        prototype._incorrectSolutionDialogShown =
            _GameWidget._incorrectSolutionDialogShown;

        prototype.__defineGetter__("_puzzleFile", _GameWidget._getPuzzleFile);
        prototype.__defineSetter__("_puzzleFile", _GameWidget._setPuzzleFile);
        prototype.__defineGetter__("puzzleFile", _GameWidget._getPuzzleFile);
        prototype.__defineSetter__("puzzleFile", _GameWidget._setPuzzleFile);

        prototype.connectSignalsAndChildren =
            _GameWidget.connectSignalsAndChildren;
        prototype.restart = _GameWidget.restart;
        prototype._acquireChildren = _GameWidget._acquireChildren;
        prototype._listColorIndices = _GameWidget._listColorIndices;
        prototype._generateClues = _GameWidget._generateClues;
        prototype._generateRowClues = _GameWidget._generateRowClues;
        prototype._generateColumnClues = _GameWidget._generateColumnClues;
        prototype._updateClueTables = _GameWidget._updateClueTables;
        prototype._updatePixelGrid = _GameWidget._updatePixelGrid;
        prototype._onPuzzleChanged = _GameWidget._onPuzzleChanged;
        prototype._onPuzzleFull = _GameWidget._onPuzzleFull;
        prototype._isPuzzleSolved = _GameWidget._isPuzzleSolved;
        prototype._onPuzzleSolved = _GameWidget._onPuzzleSolved;
        prototype._onPuzzleIncorrectSolution =
            _GameWidget._onPuzzleIncorrectSolution;
    },

    init: function() {
    },

    connectSignalsAndChildren: function(builder) {
        this._acquireChildren(builder);

        this._pixelGrid.signal.puzzle_changed.connect(
            Lang.bind(this, this._onPuzzleChanged));
        this._pixelGrid.signal.puzzle_full.connect(
            Lang.bind(this, this._onPuzzleFull));
    },

    restart: function() {
        this._pixelGrid.reset();
    },

    _acquireChildren: function(builder) {
        this._rowClueTable = builder.get_object("rowcluetable");
        this._columnClueTable = builder.get_object("columncluetable");
        this._pixelGrid = builder.get_object("pixelgrid");

        builder.add_from_file(Config.datadir + "/puzzle-end-dialog.ui");
        this._correctSolutionDialog =
            builder.get_object("correct_solution_dialog");
        this._correctSolutionDialog.transient_for = this.get_toplevel();
        this._incorrectSolutionDialog =
            builder.get_object("incorrect_solution_dialog");
        this._incorrectSolutionDialog.transient_for = this.get_toplevel();
},

    _getPuzzleFile: function() {
        return this.__puzzleFile;
    },

    _setPuzzleFile: function(value) {
        try {
            this._solutionPuzzlePixels = NonogramIO.load_puzzle(value);
        } catch (e) {
            if (e.name == "PuzzleLoadFileError")
                print("File error while loading puzzle: " + e.message);
            else if (e.name == "PuzzleLoadImageError")
                print("Image error while loading puzzle: " + e.message);
            return;
        }

        this.__puzzleFile = value;
        this._numRows = this._solutionPuzzlePixels.length;
        this._numColumns = this._solutionPuzzlePixels[0].length;

        this._listColorIndices();
        this._generateClues();

        this._updateClueTables();
        this._updatePixelGrid();

        this._incorrectSolutionDialogShown = false;
    },

    _listColorIndices: function() {
        this._colorIndices = {};
        for (var y = 0; y < this._numRows; ++y) {
            for (var x = 0; x < this._numColumns; ++x) {
                this._colorIndices[this._solutionPuzzlePixels[y][x]] = 1;
            }
        }
    },

    _generateClues: function() {
        this._generateRowClues();
        this._generateColumnClues();
    },

    _generateRowClues: function() {
        this._rowClues = [];
        for (var y = 0; y < this._numRows; ++y) {
            var singleRowClues = [];
            var currentColor = this._BACKGROUND_COLOR;
            var currentCount = 0;

            for (var x = 0; x < this._numColumns; ++x) {
                if (this._solutionPuzzlePixels[y][x] == currentColor) {
                    currentCount++;
                }
                else {
                    if (currentColor != this._BACKGROUND_COLOR)
                        singleRowClues.push({color: currentColor,
                                             count: currentCount});

                    currentColor = this._solutionPuzzlePixels[y][x];
                    currentCount = 1;
                }
            }

            if (currentColor != this._BACKGROUND_COLOR)
                singleRowClues.push({color: currentColor,
                                     count: currentCount});

            this._rowClues.push(singleRowClues);
        }
    },

    _generateColumnClues: function() {
        this._columnClues = [];
        for (var x = 0; x < this._numColumns; ++x) {
            var singleColumnClues = [];
            var currentColor = this._BACKGROUND_COLOR;
            var currentCount = 0;

            for (var y = 0; y < this._numRows; ++y) {
                if (this._solutionPuzzlePixels[y][x] == currentColor) {
                    currentCount++;
                }
                else {
                    if (currentColor != this._BACKGROUND_COLOR)
                        singleColumnClues.push({color: currentColor,
                                                count: currentCount});

                    currentColor = this._solutionPuzzlePixels[y][x];
                    currentCount = 1;
                }
            }

            if (currentColor != this._BACKGROUND_COLOR)
                singleColumnClues.push({color: currentColor,
                                        count: currentCount});

            this._columnClues.push(singleColumnClues);
        }
    },

    _updateClueTables: function() {
        this._rowClueTable.setClues(this._rowClues);
        this._columnClueTable.setClues(this._columnClues);
    },

    _updatePixelGrid: function() {
        this._pixelGrid.createBlank(this._numColumns, this._numRows);
        for (color in this._colorIndices) {
            if (color != this._BACKGROUND_COLOR) {
                this._pixelGrid.drawingColor = color;
                break;
            }
        }
    },

    _onPuzzleChanged: function() {
        var isPuzzleSolved = this._isPuzzleSolved();

        if (isPuzzleSolved)
            this._onPuzzleSolved();
    },

    _onPuzzleFull: function() {
        if (this._isPuzzleSolved() == false)
            this._onPuzzleIncorrectSolution();
    },

    _isPuzzleSolved: function() {
        var puzzlePixels = this._pixelGrid.puzzlePixels;

        for (var y = 0; y < this._numRows; ++y) {
            for (var x = 0; x < this._numColumns; ++x) {
                if (this._solutionPuzzlePixels[y][x] == this._BACKGROUND_COLOR
                    && puzzlePixels[y][x] == this._NO_COLOR)
                    continue;

                if (this._solutionPuzzlePixels[y][x] != puzzlePixels[y][x])
                    return false;
            }
        }

        return true;
    },

    _onPuzzleSolved: function() {
        this._correctSolutionDialog.run();
        this._correctSolutionDialog.hide();
        // TODO: disable puzzle interface
    },

    _onPuzzleIncorrectSolution: function() {
        if (this._incorrectSolutionDialogShown)
            return;

        this._incorrectSolutionDialogShown = true;

        this._incorrectSolutionDialog.run();
        this._incorrectSolutionDialog.hide();
    }
});
